#include <windows.h>
#include <process.h>
#include <shlwapi.h>
#include <cstdio>
#include <string>
#include <vector>
#include <cassert>

class CCriticalSection
{
public:
    CRITICAL_SECTION m_csLock;

    CCriticalSection()
    {
        ::InitializeCriticalSection(&m_csLock);
    }

    virtual ~CCriticalSection()
    {
        ::DeleteCriticalSection(&m_csLock);
    }

    void Lock()
    {
        ::EnterCriticalSection(&m_csLock);
    }

    void Unlock()
    {
        ::LeaveCriticalSection(&m_csLock);
    }
};

class CObj
{
public:
    CObj();
    virtual ~CObj();
    ULONG AddRef();
    ULONG Release();
    LONG GetRefCount() const;

#ifndef NDEBUG
    static LONG s_nAliveCount;
#endif
protected:
    LONG m_cRefs;
};

class CDirChangeHandler : public CObj
{
public:
    CDirChangeHandler();
    virtual ~CDirChangeHandler();
    virtual BOOL OnChangeNotify(DWORD dwAction, LPCWSTR file1, LPCWSTR file2);
};

class CDirWatchItem : public CObj
{
private:
    CDirWatchItem() = delete;
    CDirWatchItem(const CDirWatchItem&) = delete;
    CDirWatchItem& operator=(const CDirWatchItem&) = delete;

public:
    std::wstring m_strDir;
    HANDLE m_hDir;
    DWORD m_dwChangeFilter;
    BOOL m_bWatchSubDir;
    DWORD m_dwReadDirError;
    HANDLE m_hStartEvent;
    OVERLAPPED m_Overlapped;

    CDirWatchItem(LPCWSTR pszDir, DWORD dwChangeFilter, BOOL bWatchSubDir);
    virtual ~CDirWatchItem();

    BOOL ReopenDir(HANDLE& hCompPort);
    BOOL IsDir();
    BOOL WaitForStartEvent();
    void SetStartEvent();

protected:
    void CloseDirHandle();
    void CloseStartEvent();
};

enum WATCH_STATE
{
    WATCH_STATE_INITIAL,
    WATCH_STATE_ADD_ITEM,
    WATCH_STATE_WAIT,
    WATCH_STATE_STOP
};

class CDirChangeWatcher : public CObj, public CCriticalSection
{
public:
    CDirChangeWatcher();
    virtual ~CDirChangeWatcher();

    void SetChangeHandler(CDirChangeHandler *pChangeHandler);

    BOOL IsWatching();
    BOOL IsWatchingDir(LPCWSTR pszDir);
    DWORD WatchDir(LPCWSTR pszDir, DWORD dwChangeFilter, BOOL bWatchSubDir);
    BOOL UnwatchDir(LPCWSTR pszDir);
    BOOL UnwatchDir(CDirWatchItem *pWatchItem);
    BOOL UnwatchDir(INT_PTR iItem);
    BOOL UnwatchAll();

    INT_PTR FindWatchItem(LPCWSTR pszDir) const;
    INT_PTR FindWatchItemExact(LPCWSTR pszDir) const;
    INT_PTR FindWatchItem(CDirWatchItem *pWatchItem) const;
    void ReopenDirs();

protected:
    HANDLE m_hCompPort;
    HANDLE m_hThread;
    unsigned m_dwThreadID;
    CDirChangeHandler *m_pChangeHandler;
    WATCH_STATE m_nWatchState;
    std::vector<CDirWatchItem*> m_WatchItems;

    DWORD WatchProc();
    BOOL WaitForShutdown(CDirWatchItem *pWatchItem);
    INT_PTR AddWatchItem(CDirWatchItem *pWatchItem);
    void CloseCompPort();
    void CloseThread();
    void ReleaseChangeHandler();
    void ProcessNotifications(CDirWatchItem *pWatchItem, LPBYTE pbBuffer, size_t cbBuffer);
    BOOL IsThreadRunning();

    static unsigned __stdcall WatchThreadProc(void *pThis)
    {
        CDirChangeWatcher *pWatcher = reinterpret_cast<CDirChangeWatcher *>(pThis);
        assert(pWatcher);
        return pWatcher->WatchProc();
    }
};
